#line 1513 "ExpandModules.nw"
#include "ExpandModules.hpp"

namespace ExpandModules {
  using namespace IR;
  using namespace AST;
  
#line 150 "ExpandModules.nw"
const set<Module*> *FindRoots::find()
{
  // Start the set of roots off as all modules
  for ( vector<Module*>::const_iterator i = modules->modules.begin() ;
	i != modules->modules.end() ; i++ ) {
    assert(*i);
    roots.insert(*i);
  }

  // Visit each of the modules and remove any run modules from the roots set

  for ( vector<Module*>::const_iterator i = modules->modules.begin() ;
	i != modules->modules.end() ; i++ )
    recurse(*i);

  return &roots;
}
#line 186 "ExpandModules.nw"
Status FindRoots::visit(Run &s)
{
  if ( !(modules->module_symbols.contains(s.old_name)) )
    throw IR::Error("Attempt to run unknown module " + s.old_name);
  ModuleSymbol *ms =
    dynamic_cast<ModuleSymbol*>(modules->module_symbols.get(s.old_name));
  assert(ms);
  Module *m = ms->module;
  assert(m);
  roots.erase(m);
  return Status();
}
#line 1519 "ExpandModules.nw"
  
#line 442 "ExpandModules.nw"
void Copier::copySymbolTable(const SymbolTable *source, SymbolTable *dest)
{
  assert(source);
  assert(dest);

  for ( SymbolTable::const_iterator i = source->begin() ;
        i != source->end() ; i++) {
    assert(*i);
    Symbol *news = copy(*i);
    dest->enter(news);
    copiedSymbolMap.insert( std::make_pair(*i, news) );
    // std::cout << "Copied " << news->name << std::endl;
  }

  if (source->parent)
    dest->parent = newSymbolTable(source->parent);

  copiedSymbolTableMap.insert( std::make_pair(source, dest) );
}
#line 475 "ExpandModules.nw"
Module *Copier::copyModule(Module &m)
{
  assert(m.symbol);
  ModuleSymbol *ms = new ModuleSymbol(m.symbol->name);
  Module *result = new Module(ms);
  newModule = result;
  ms->module = result;

  // std::cout << "Copying " << m.symbol->name << std::endl;

  copySymbolTable(m.types, result->types);
  copySymbolTable(m.constants, result->constants);
  copySymbolTable(m.functions, result->functions);
  copySymbolTable(m.procedures, result->procedures);
  copySymbolTable(m.tasks, result->tasks);
  copySymbolTable(m.variables, result->variables);
  copySymbolTable(m.signals, result->signals);

  for ( vector<InputRelation*>::const_iterator i = m.relations.begin() ;
	i != m.relations.end() ; i++ ) {
    assert(*i);
    result->relations.push_back(copy(*i));
  }

  for ( vector<Counter*>::const_iterator i = m.counters.begin() ;
	i != m.counters.end() ; i++ ) {
    assert(*i);
    result->counters.push_back(*i);
    copiedCounterMap[*i] = *i;
  }

  result->body = copy(m.body);
  return result;  
}
#line 541 "ExpandModules.nw"
Status Copier::visit(Run &r)
{
  assert(newModule);

  // std::cout << "run " << r.new_name << " / " << r.old_name << std::endl;

  // Locate the old module by its name
  assert(oldModules);
  if ( !(oldModules->module_symbols.contains(r.old_name)) )
    throw IR::Error("Attempt to run unknown module " + r.old_name);

  ModuleSymbol *ms =
    dynamic_cast<ModuleSymbol*>(oldModules->module_symbols.get(r.old_name));
  assert(ms);
  Module *moduleToCopy = ms->module;
  assert(moduleToCopy);

  Statement *body = dynamic_cast<Statement*>(moduleToCopy->body);
  assert(body);

  Copier newcopier(this);

  // Map the declaration symbol tables in the module being run to the
  // equivalent ones in the new module
  newcopier.copiedSymbolTableMap.insert(std::make_pair( moduleToCopy->constants,
                                                        newModule->constants));

  
#line 608 "ExpandModules.nw"
  for ( vector<SignalRenaming*>::const_iterator i = r.signals.begin() ;
	i != r.signals.end() ; i++ ) {
    SignalRenaming *sr = *i;
    assert(sr);
    if ( !(moduleToCopy->signals->local_contains(sr->old_name)) )
      throw IR::Error("Attempting to rename unknown signal " + sr->old_name);
    SignalSymbol *oldSignalSymbol =
      dynamic_cast<SignalSymbol*>(moduleToCopy->signals->get(sr->old_name));
    assert(oldSignalSymbol);
    SignalSymbol *newSignalSymbol = newSymbol(sr->new_sig);
    newcopier.copiedSymbolMap.insert(std::make_pair(oldSignalSymbol,
						    newSignalSymbol));
  }
#line 569 "ExpandModules.nw"
  
#line 631 "ExpandModules.nw"
  assert(r.signalScope); // Should have been put there during static semantics
  SymbolTable *newSignalTable = newSymbolTable(r.signalScope);
  newcopier.copiedSymbolTableMap.insert(std::make_pair(moduleToCopy->signals,
	  					       newSignalTable));

  for ( SymbolTable::const_iterator i = moduleToCopy->signals->begin() ;
	i != moduleToCopy->signals->end() ; i++ ) {    
    SignalSymbol *oldSignal = dynamic_cast<SignalSymbol*>(*i);
    assert(oldSignal);
    if ( newcopier.copiedSymbolMap.find(oldSignal) ==
	 newcopier.copiedSymbolMap.end() ) {
      if (!newSignalTable->contains(oldSignal->name))
	throw IR::Error("Could not find signal " + oldSignal->name +
			" in enclosing scope");     
      SignalSymbol *newSignal =
	dynamic_cast<SignalSymbol*>(newSignalTable->get(oldSignal->name));
      assert(newSignal);
      newcopier.copiedSymbolMap.insert( std::make_pair(oldSignal, newSignal) );
    }
  }

#line 571 "ExpandModules.nw"
  
#line 664 "ExpandModules.nw"
  for ( vector<TypeRenaming*>::const_iterator i = r.types.begin() ;
	i != r.types.end() ; i++ ) {
    TypeRenaming *tr = *i;
    assert(tr);
    if ( !(moduleToCopy->types->local_contains(tr->old_name)) )
      throw IR::Error("Attempting to rename unknown type " + tr->old_name);
    TypeSymbol *oldType =
      dynamic_cast<TypeSymbol*>(moduleToCopy->types->get(tr->old_name));
    assert(oldType);
    TypeSymbol *newType = newSymbol(tr->new_type);
    newcopier.copiedSymbolMap.insert(std::make_pair(oldType, newType));
  }
#line 572 "ExpandModules.nw"
  
#line 684 "ExpandModules.nw"
#if 0
  std::cout << "Start of symbol table" << std::endl;
  for ( SymbolTable::const_iterator i = newModule->types->begin() ;
	i != newModule->types->end() ; i++ ) {
    assert(*i);
    std::cout << "type " << (*i)->name << std::endl;
  }
  std::cout << "End of symbol table" << std::endl;
#endif

  for ( SymbolTable::const_iterator i = moduleToCopy->types->begin() ;
	i != moduleToCopy->types->end() ; i++ ) {
    TypeSymbol *oldType = dynamic_cast<TypeSymbol*>(*i);
    assert(oldType);
    // std::cout << "copying type " << oldType->name << std::endl;

    if ( newcopier.copiedSymbolMap.find(oldType) ==
	 newcopier.copiedSymbolMap.end() ) {

      TypeSymbol *newType;      
      if ( newModule->types->local_contains(oldType->name) ) {
	// Use existing type
	newType = dynamic_cast<TypeSymbol*>(newModule->types->get(oldType->name));
	assert(newType);
      } else {
	// Copy type into running module
	newType = newcopier.copy(oldType);
	newModule->types->enter(newType);
      }

      assert(newType);
      newcopier.copiedSymbolMap.insert( std::make_pair(oldType, newType) );
    }
  }

#line 574 "ExpandModules.nw"
  
#line 726 "ExpandModules.nw"
for ( vector<ConstantRenaming*>::const_iterator i = r.constants.begin() ;
	i != r.constants.end() ; i++ ) {
  ConstantRenaming *cr = *i;
  assert(cr);
  if ( !(moduleToCopy->constants->local_contains(cr->old_name)) )
    throw IR::Error("Attempting to rename unknown constant " + cr->old_name);
  ConstantSymbol *oldConstant =
    dynamic_cast<ConstantSymbol*>(moduleToCopy->constants->get(cr->old_name));
  assert(oldConstant);
  ExpressionCopier *ec = new ExpressionCopier(cr->new_value, this);
  newcopier.newConstantExpression.insert(std::make_pair(oldConstant, ec));
}
#line 575 "ExpandModules.nw"
  
#line 741 "ExpandModules.nw"
for ( SymbolTable::const_iterator i = moduleToCopy->constants->begin() ;
	i != moduleToCopy->constants->end() ; i++ ) {
  ConstantSymbol *oldConstant = dynamic_cast<ConstantSymbol*>(*i);
  assert(oldConstant);
  // Here, we have a constant in the instantiated module.  It may be that
  // 1. The constant is new: add it to the module being built
  // 2. The constant already exists in the main module: use that one
  // 3. The constant has been renamed: do nothing; this was handled above

  if ( newcopier.newConstantExpression.find(oldConstant) ==
	 newcopier.newConstantExpression.end() ) {

    // Have not already renamed this constant

    ConstantSymbol *newConstant;

    if ( newModule->constants->local_contains(oldConstant->name) ) {
	// Found the constant: use the existing one
	newConstant =
         dynamic_cast<ConstantSymbol*>(newModule->constants->get(oldConstant->name));
      /* FIXME: Should verify that the initial value of the two
	   constants is consistent

	   module Foo:
	   constant c = 10 : integer;
	   run Bar
	   end module

	   module Bar:
	   constant c = 20 : integer;
	   nothing
	   end module
	*/

    } else {
	// Constant doesn't already exist in the new module: copy it
	// std::cout << "Making a new copy of " << oldConstant->name << std::endl; 
	newConstant = newcopier.copy(oldConstant);
	newModule->constants->enter(newConstant);

	/* FIXME: This won't generate re-parsable code if the
	   constant's name was hidden by a local variable decalaration, e.g.,

	   module Foo:
	   var c : integer in
	     run Bar
	   end
           end module

           module Bar:
           constant c : integer;
           nothing
           end module

	   Berry's iclc doesn't worry about this because it simply numbers
	   everything and stores the names separately.
	*/
    }

    assert(newConstant);
    newcopier.copiedSymbolMap.insert( std::make_pair(oldConstant, newConstant) );
  }
}

#line 577 "ExpandModules.nw"
  
#line 809 "ExpandModules.nw"
/* FIXME: Renaming built-in functions, such as +, will generate
   strange output,

   e.g., c := +(a,b)
 */

for ( vector<FunctionRenaming*>::const_iterator i = r.functions.begin() ;
	i != r.functions.end() ; i++ ) {
  FunctionRenaming *fr = *i;
  assert(fr);
  if ( !(moduleToCopy->functions->local_contains(fr->old_name)) )
    throw IR::Error("Attempting to rename unknown function " + fr->old_name);
  FunctionSymbol *oldFunction =
    dynamic_cast<FunctionSymbol*>(moduleToCopy->functions->get(fr->old_name));
  assert(oldFunction);
  FunctionSymbol *newFunction = newSymbol(fr->new_func);
  newcopier.copiedSymbolMap.insert(std::make_pair(oldFunction, newFunction));
}
#line 578 "ExpandModules.nw"
  
#line 830 "ExpandModules.nw"
for ( SymbolTable::const_iterator i = moduleToCopy->functions->begin() ;
	i != moduleToCopy->functions->end() ; i++ ) {
  FunctionSymbol *oldFunction = dynamic_cast<FunctionSymbol*>(*i);
  assert(oldFunction);
  // Here, we have a function in the instantiated module.  It may be that
  // 1. The function is new: add it to the module being built
  // 2. The function already exists in the main module: use that one
  // 3. The function has been renamed: do nothing; this was handled above

  if ( newcopier.copiedSymbolMap.find(oldFunction) ==
	 newcopier.copiedSymbolMap.end() ) {

    // Haven't already renamed this function

    FunctionSymbol *newFunction;

    if ( newModule->functions->local_contains(oldFunction->name) ) {
	// Found the function: use the existing one
	newFunction =
         dynamic_cast<FunctionSymbol*>(newModule->functions->get(oldFunction->name));

	// FIXME: Should check that the parameter types are consistent

    } else {
	// Function doesn't already exist in the new module: copy it
	newFunction = newcopier.copy(oldFunction);
	newModule->functions->enter(newFunction);
    }

    assert(newFunction);
    newcopier.copiedSymbolMap.insert( std::make_pair(oldFunction, newFunction) );
  }
}

#line 580 "ExpandModules.nw"
  
#line 871 "ExpandModules.nw"
for ( vector<ProcedureRenaming*>::const_iterator i = r.procedures.begin() ;
	i != r.procedures.end() ; i++ ) {
  ProcedureRenaming *pr = *i;
  assert(pr);
  if ( !(moduleToCopy->procedures->local_contains(pr->old_name)) )
    throw IR::Error("Attempting to rename unknown procedure " + pr->old_name);
  ProcedureSymbol *oldProcedure =
    dynamic_cast<ProcedureSymbol*>(moduleToCopy->procedures->get(pr->old_name));
  assert(oldProcedure);
  ProcedureSymbol *newProcedure = newSymbol(pr->new_proc);
  newcopier.copiedSymbolMap.insert(std::make_pair(oldProcedure, newProcedure));
}
#line 581 "ExpandModules.nw"
  
#line 896 "ExpandModules.nw"
for ( SymbolTable::const_iterator i = moduleToCopy->procedures->begin() ;
	i != moduleToCopy->procedures->end() ; i++ ) {
  ProcedureSymbol *oldProcedure = dynamic_cast<ProcedureSymbol*>(*i);
  assert(oldProcedure);

  if ( newcopier.copiedSymbolMap.find(oldProcedure) ==
	 newcopier.copiedSymbolMap.end() ) {

    // Haven't already renamed this procedure
    ProcedureSymbol *newProcedure;

    if ( newModule->procedures->local_contains(oldProcedure->name) ) {
	// Found the procedure: use the existing one
	newProcedure =
         dynamic_cast<ProcedureSymbol*>(newModule->procedures->get(oldProcedure->name));

	// FIXME: Verify the parameter count and types are consistent

    } else {
	// Procedure doesn't already exist in the new module: copy it
	newProcedure = newcopier.copy(oldProcedure);
	newModule->procedures->enter(newProcedure);
    }
    assert(newProcedure);
    newcopier.copiedSymbolMap.insert( std::make_pair(oldProcedure, newProcedure) );
  }
}

#line 583 "ExpandModules.nw"
  
#line 930 "ExpandModules.nw"
for ( vector<ProcedureRenaming*>::const_iterator i = r.tasks.begin() ;
	i != r.tasks.end() ; i++ ) {
  ProcedureRenaming *pr = *i;
  assert(pr);
  if ( !(moduleToCopy->tasks->local_contains(pr->old_name)) )
    throw IR::Error("Attempting to rename unknown task " + pr->old_name);
  TaskSymbol *oldTask =
    dynamic_cast<TaskSymbol*>(moduleToCopy->tasks->get(pr->old_name));
  assert(oldTask);
  TaskSymbol *newTaskSymbol = dynamic_cast<TaskSymbol*>(pr->new_proc);
  assert(newTaskSymbol);
  TaskSymbol *newTask = newSymbol(newTaskSymbol);
  newcopier.copiedSymbolMap.insert(std::make_pair(oldTask, newTask));
}
#line 584 "ExpandModules.nw"
  
#line 947 "ExpandModules.nw"
for ( SymbolTable::const_iterator i = moduleToCopy->tasks->begin() ;
	i != moduleToCopy->tasks->end() ; i++ ) {
  TaskSymbol *oldTask = dynamic_cast<TaskSymbol*>(*i);
  assert(oldTask);

  if ( newcopier.copiedSymbolMap.find(oldTask) ==
	 newcopier.copiedSymbolMap.end() ) {

    // Haven't already renamed this task
    TaskSymbol *newTask;

    if ( newModule->tasks->local_contains(oldTask->name) ) {
	// Found the task: use the existing one
	newTask =
         dynamic_cast<TaskSymbol*>(newModule->tasks->get(oldTask->name));

	// FIXME: Verify the parameter count and types are consistent

    } else {
	// Task doesn't already exist in the new module: copy it
	newTask = newcopier.copy(oldTask);
	newModule->tasks->enter(newTask);
    }
    assert(newTask);
    newcopier.copiedSymbolMap.insert( std::make_pair(oldTask, newTask) );
  }
}

#line 586 "ExpandModules.nw"
  
#line 985 "ExpandModules.nw"
for ( SymbolTable::const_iterator i = moduleToCopy->variables->begin() ;
	i != moduleToCopy->variables->end() ; i++ ) {
  VariableSymbol *oldVariable = dynamic_cast<VariableSymbol*>(*i);
  assert(oldVariable);
  string baseName = r.new_name + "_" + oldVariable->name;
  string newName = baseName;
  int next = 1;
  while (newModule->variables->local_contains(newName)) {
    char buf[10];
    sprintf(buf, "%d", next++);
    newName = baseName + '_' + buf;
  }
  TypeSymbol *newType = newcopier.newSymbol(oldVariable->type);
  Expression *newInitializer = newcopier.copy(oldVariable->initializer);
  VariableSymbol *newVariable =
     new VariableSymbol(newName, newType, newInitializer);
  newModule->variables->enter(newVariable);
  newcopier.copiedSymbolMap.insert( std::make_pair(oldVariable, newVariable));
//  std::cout << "copying " << oldVariable->name << " to " << newName << std::endl;
}

copiedSymbolTableMap.insert( std::make_pair(moduleToCopy->variables,
	                                    newModule->variables) );

#line 588 "ExpandModules.nw"
  
#line 1017 "ExpandModules.nw"
for ( vector<Counter*>::const_iterator i = moduleToCopy->counters.begin() ;
	i != moduleToCopy->counters.end() ; i++ ) {
  Counter *newCounter = new Counter();
  newModule->counters.push_back(newCounter);
  newcopier.copiedCounterMap[*i] = newCounter;
}

#line 590 "ExpandModules.nw"
  Statement *result = newcopier.copy(body);
  return result;
}
#line 1053 "ExpandModules.nw"
Status Copier::visit(FunctionSymbol &s)
{
  FunctionSymbol *result = new FunctionSymbol(s.name);
  result->result = newSymbol(s.result);
  for ( vector<TypeSymbol*>::const_iterator i = s.arguments.begin() ;
        i != s.arguments.end() ; i++ ) {
    assert(*i);
    result->arguments.push_back(newSymbol(*i));
  }
  return result;
}

Status Copier::visit(BuiltinFunctionSymbol &s)
{
  BuiltinFunctionSymbol *result = new BuiltinFunctionSymbol(s.name);
  result->result = newSymbol(s.result);
  for ( vector<TypeSymbol*>::const_iterator i = s.arguments.begin() ;
        i != s.arguments.end() ; i++ ) {
    assert(*i);
    result->arguments.push_back(newSymbol(*i));
  }
  return result;
}
#line 1079 "ExpandModules.nw"
Status Copier::visit(ProcedureSymbol &s)
{
  ProcedureSymbol *result = new ProcedureSymbol(s.name);
  for ( vector<TypeSymbol*>::const_iterator i = s.reference_arguments.begin() ;
        i != s.reference_arguments.end() ; i++ ) {
    assert(*i);
    result->reference_arguments.push_back(newSymbol(*i));
  }
  for ( vector<TypeSymbol*>::const_iterator i = s.value_arguments.begin() ;
        i != s.value_arguments.end() ; i++ ) {
    assert(*i);
    result->value_arguments.push_back(newSymbol(*i));
  }
  return result;
}

Status Copier::visit(TaskSymbol &s)
{
  TaskSymbol *result = new TaskSymbol(s.name);
  for ( vector<TypeSymbol*>::const_iterator i = s.reference_arguments.begin() ;
        i != s.reference_arguments.end() ; i++ ) {
    assert(*i);
    result->reference_arguments.push_back(newSymbol(*i));
  }
  for ( vector<TypeSymbol*>::const_iterator i = s.value_arguments.begin() ;
        i != s.value_arguments.end() ; i++ ) {
    assert(*i);
    result->value_arguments.push_back(newSymbol(*i));
  }
  return result;
}
#line 1176 "ExpandModules.nw"
Status Copier::visit(ProcedureCall &c)
{
  ProcedureCall *result = new ProcedureCall(newSymbol(c.procedure));
  for ( vector<VariableSymbol*>::const_iterator i = c.reference_args.begin() ;
	i != c.reference_args.end() ; i++ ) {
    assert(*i);
    result->reference_args.push_back(newSymbol(*i));
  }
  for ( vector<Expression*>::const_iterator i = c.value_args.begin() ;
	i != c.value_args.end() ; i++ ) {
    assert(*i);
    result->value_args.push_back(copy(*i));
  }
  return result;
}
#line 1203 "ExpandModules.nw"
Status Copier::visit(StatementList &l)
{
  StatementList *result = new StatementList();
  for (vector<Statement*>::iterator i = l.statements.begin() ;
       i != l.statements.end() ; i++ ) {
    *result << copy(*i);
  }
  return result;
}

Status Copier::visit(ParallelStatementList &l)
{
  ParallelStatementList *result = new ParallelStatementList();
  for (vector<Statement*>::iterator i = l.threads.begin() ;
       i != l.threads.end() ; i++ ) {
    result->threads.push_back(copy(*i));
  }
  return result;
}
#line 1279 "ExpandModules.nw"
Status Copier::visit(Exec &e)
{
  Exec *result = new Exec();
  for ( vector<TaskCall*>::const_iterator j = e.calls.begin() ;
	j != e.calls.end() ; j++ ) {
    TaskCall *c = *j;
    assert(c);
    TaskSymbol *ts = dynamic_cast<TaskSymbol*>(c->procedure);
    assert(ts);
    TaskCall *newCall = new TaskCall(newSymbol(ts));
    newCall->signal = newSymbol(c->signal);
    newCall->body = copy(c->body);
    for ( vector<VariableSymbol*>::const_iterator i = c->reference_args.begin() ;
	  i != c->reference_args.end() ; i++ ) {
      assert(*i);
      newCall->reference_args.push_back(newSymbol(*i));
    }
    for ( vector<Expression*>::const_iterator i = c->value_args.begin() ;
	  i != c->value_args.end() ; i++ ) {
      assert(*i);
      newCall->value_args.push_back(copy(*i));
    }
    result->calls.push_back(newCall);
  }
  return result;
}
#line 1315 "ExpandModules.nw"
void Copier::copyCases(const CaseStatement &source, CaseStatement *dest)
{  
  assert(dest);
  for ( vector<PredicatedStatement*>::const_iterator i = source.cases.begin() ;
	i != source.cases.end() ; i++) {
    assert(*i);
    dest->newCase(copy((*i)->body), copy((*i)->predicate));
  }

  dest->default_stmt = copy(source.default_stmt);
}
#line 1473 "ExpandModules.nw"
Status Copier::visit(FunctionCall &c)
{
  FunctionCall *result = new FunctionCall(newSymbol(c.callee));
  for ( vector<Expression*>::const_iterator i = c.arguments.begin() ;
        i != c.arguments.end() ; i++ ) {
    assert(*i);
    result->arguments.push_back(copy(*i));
  }
  return result;
}
#line 1520 "ExpandModules.nw"
  
#line 1421 "ExpandModules.nw"
ExpressionCopier::ExpressionCopier(Expression *e, Copier *c)
  : theExpr(e), theCopier(c) { assert(e); assert(c); }

Expression *ExpressionCopier::copy() { return theCopier->copy(theExpr); }
#line 1521 "ExpandModules.nw"
  
#line 94 "ExpandModules.nw"
Modules *rewrite(Modules *m)
{
  assert(m);
  FindRoots fr(m);
  const set<Module*> *roots = fr.find();
  if (roots->empty())
    throw IR::Error("No root nodes.  Is there a cyclic module dependency?");

  // Copy each root module in the order in which it appeared
  // in the original list of modules (preserve order)
  Modules *result = new Modules();
  for ( vector<Module*>::const_iterator i = m->modules.begin() ;
	i != m->modules.end() ; i++ )
    if ( roots->find(*i) != roots->end() ) {
      Copier copier(m, result);
      result->add(copier.copyModule(**i));
    }
  
  return result;
}
#line 1522 "ExpandModules.nw"
}
