#include "IR.hpp"
#include "AST.hpp"
#include "GRCPrinter.hpp"
#include "CPrinter.hpp"
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <set>

using namespace AST;
using namespace std;
using namespace GRCDot;

class BITMAP{
public:
  unsigned int *bitmap;
  int sz, nwords;

  BITMAP(int sz): sz(sz), nwords(1 + (sz-1)/32)
  {
    bitmap = new unsigned int [nwords];
  }

  BITMAP(int sz, bool val): sz(sz), nwords(1 + (sz-1)/32)
  {
    bitmap = new unsigned int [nwords];

    if (val)
      for(int i = 0; i < nwords; i++){
	bitmap[i] = 0xffffffff;
      }
    else
       for(int i = 0; i < nwords; i++){
	bitmap[i] = 0;
       }
  }
  
  void And(BITMAP *in)
  {
    for(int i = 0; i< nwords; i++)
      bitmap[i] &= in->bitmap[i];
  }

  void Or(BITMAP *in)
  {
    for(int i = 0; i< nwords; i++)
      bitmap[i] |= in->bitmap[i];
  }

  void SetBit(int pos, unsigned int val)
  {
    int ipart,lpart;
    assert(pos < sz);
    ipart = pos / 32;
    lpart = pos % 32;
    bitmap[ipart] |= (val<<lpart);
  }

  int GetBit(int bit)
  {
    int RET;
    RET = (bitmap[bit/32]>>(bit%32)) & 0x1;
    return RET;
  }

  BITMAP *copy()
  {
    BITMAP *RET = new BITMAP(sz,0);
    for (int i = 0; i < nwords; i++)
      RET->bitmap[i] |= bitmap[i];
    return RET;
  }

  void print() //for debug
  {
    unsigned int bitval;

    //cerr<<"BITMAP = ";
    for (int i = 0; i < sz; i++){
      bitval = (bitmap[i/32]>>(i%32)) & 0x1;
      cerr<<bitval;
    }
    cerr<<"\n";
  }
};

class Edge
{
public:
  GRCNode *begin;
  GRCNode *end;
  Edge(GRCNode *begin, GRCNode *end):begin(begin),end(end){}
};

class ExEdge
{
public:
  //GRCNode *head;
  set<GRCNode *> descends;
  set<Edge *> exedges;

  //ExEdge(GRCNode *head):head(head)
  ExEdge()
  {
  }

  void AddDesc(GRCNode *n)
  {
    if(n)
      descends.insert(n);
  }

  void AddDesc(set<GRCNode *> nset)
  {
    descends.insert(nset.begin(), nset.end());   
  }

  void FindExedges(GRCNode *head)
  {
    set<GRCNode *>::iterator i;
    vector<GRCNode *>::iterator j;
    for(i = descends.begin(); i != descends.end(); i++){
      if((*i) == head)
	continue;
      for (j = (*i)->predecessors.begin(); j != (*i)->predecessors.end(); j++){
	assert(*j);
	if (descends.count(*j)==0){
	  Edge *edge = new Edge(*j,*i);
	  exedges.insert(edge);
	}
      }
    }
  }

};

class EEC{

  GRCNode *top;
  SymbolTable *sigs;
  GRCNode::NumMap &refmap; 
  GRCNode *toppdg;
  EnterGRC *entergrc;

  set<GRCNode *> visited;
  set<GRCNode *> par_visit;
  vector<GRCNode *> topo;
  map<GRCNode *, int> bitpos;
  map<GRCNode *, BITMAP *> region;
  map<GRCNode *, BITMAP *> eec;
  map<GRCNode *, set<GRCNode *> > dpsnodes;
  map<GRCNode *, vector<GRCNode *> > conn_pred;
  map<GRCNode *, vector<GRCNode *> > conn_succ;
  map<GRCNode *, GRCNode *> start; //start of a subgraph under n
  map<GRCNode *, vector<GRCNode *> > ends; //ends of a subgraph under n
  
  int debug;
  int pos_cnt;
  int var_cnt; //counter for naming new guard-variable when making cut
  int mxnode;

public:
  EEC(GRCNode *top, SymbolTable *sigs, CFGmap &refmap, int mxnode): top(top), sigs(sigs), refmap(refmap), mxnode(mxnode)
  {
    debug = 0;
    
    entergrc = dynamic_cast<EnterGRC*>(top);
    assert(entergrc);
    assert(entergrc->successors.size() == 1);
    toppdg = dynamic_cast<GRCNode*>(entergrc ->successors[0]); 
    assert (toppdg);
  }

  void excute()
  {   
    DFS(entergrc);
    
    //due to Ferrante's alg, only fork nodes may have multi-parents
    ParForking();

    visited.clear();
    pos_cnt = 0;
    DFSCounter(entergrc);
    //printTopo();

    ComputeRegion();

    visited.clear();
    DFSEec(entergrc);
    //BitmapPrint();

    var_cnt = 0;
    OrderChildren();

    visited.clear();
    ReplaceNull2Nop(entergrc);

    RmLinks();

    visited.clear();    
    DFSRemoveForks(entergrc);

    visited.clear();
    ReplaceNop2Null(entergrc);
    
    //GRCNode *exitgrc = addExitGRC();
    //cfg_top = rmEnterGRC();
    //return exitgrc;
    
  }

  GRCNode *rtTop()
  {   
    GRCNode *org_top = top;
    assert(top->successors.size()>0);
    top = top->successors[0];
    assert(top != org_top);
    org_top->successors.clear();    
    top->predecessors.clear();
    assert(top->successors.size()>0);
    return top;
  }

  GRCNode *addExitGRC()
  {
    set<GRCNode *>::iterator i;
    set<GRCNode *> end_nodes;
    GRCNode *exitgrc = new ExitGRC();

    for (i = visited.begin(); i != visited.end(); i++){

      if ((*i)->successors.size() == 0){
	if (((*i)->predecessors.size() > 0) || (dynamic_cast<EnterGRC *>(*i)))
	  end_nodes.insert(*i);
      }
    }
    
    for (i = end_nodes.begin(); i != end_nodes.end(); i++){
      if (dynamic_cast<Nop *>(*i)){
	assert((*i)->predecessors.size() == 1);
	exitgrc->predecessors.push_back((*i)->predecessors[0]);
	vector<GRCNode *>::iterator j;
	GRCNode *par = (*i)->predecessors[0];
	for (j = par->successors.begin(); j != par->successors.end(); j++){
	  if (*j == *i){
	    *j = exitgrc;
	    break;
	  }
	}
      }
      else{
	(*i)->successors.push_back(exitgrc);
	exitgrc->predecessors.push_back(*i);
      }
    }

    return exitgrc;
  }

  void ReplaceNull2Nop(GRCNode *n)
  {
    assert(n);

    if (visited.find(n) != visited.end())
      return;
    visited.insert(n);
    
    vector<GRCNode *>::iterator i;
    for (i = conn_succ[n].begin(); i != conn_succ[n].end(); i++)
      if (!*i){
	*i = new Nop();
	conn_pred[*i].push_back(n);
	visited.insert(*i);
      }
      else
	ReplaceNull2Nop(*i);
  }

  void ReplaceNop2Null(GRCNode *n)
  {
    if (visited.find(n) != visited.end())
      return;
    visited.insert(n);
    
    vector<GRCNode *>::iterator i;
    for (i = n->successors.begin(); i != n->successors.end(); i++)
      if (dynamic_cast<Nop *>(*i)){
	assert((*i)->successors.size() == 0);
	*i = NULL;
      }
      else
	ReplaceNop2Null(*i);
  }

  void printTopo()
  {
    vector<GRCNode *>::iterator i;

    cerr<<"TOPO: ";
    for (i = topo.begin(); i != topo.end(); i++)
      cerr<<refmap[*i]<<" ";
    cerr<<"\n";
  }

  void DFSCounter(GRCNode *n)
  {
    vector<GRCNode *>::iterator i;

    if(!n) return;
    
    if(visited.count(n)>0) 
      return;
    visited.insert(n);

    bitpos[n] = pos_cnt;
    pos_cnt++;
    for (i = n->successors.begin(); i != n->successors.end(); i++){
      DFSCounter(*i);
      if(dpsnodes.count(*i) > 0)
	dpsnodes[n].insert(dpsnodes[*i].begin(), dpsnodes[*i].end());
    }

    if ((n->dataSuccessors.size()>0)||(n->dataPredecessors.size()>0))
      dpsnodes[n].insert(n);

    topo.insert(topo.begin(),n);
  }

  void topoDFS(GRCNode *n)
  {
    vector<GRCNode *>::iterator i;
    
    if(!n) return;
    
    if(visited.count(n)>0) 
      return;
    
    visited.insert(n);
    
    for (i = n->successors.begin(); i != n->successors.end(); i++){
      topoDFS(*i);
    }
    
    if (n->dataSuccessors.size()>0){
      for (i = n->dataSuccessors.begin();
	   i != n->dataSuccessors.end(); i++)
	topoDFS(*i);
    }
    
    topo.insert(topo.begin(),n);
    
    return;
  }

  void DFS(GRCNode *n)
  {
    vector<GRCNode *>::iterator i;

    if(!n) return;
    
    if(visited.count(n)>0) 
      return;
    visited.insert(n);
    for (i = n->successors.begin(); i != n->successors.end(); i++)
      DFS(*i);
  }

  void ParForking()
  {
    set<GRCNode *>::iterator i;

    for(i = visited.begin(); i != visited.end(); i++){
      //if ((*i)->successors.size() == 0){
      //assert(TypeCheck(*i)==1);
      if (TypeCheck(*i) != 3){
	if((*i)->predecessors.size()>1){
	  //cerr<<"parents forking: N"<<refmap[*i];
	  GRCNode *new_fork = new Fork();
	  visited.insert(new_fork);
	  InsertPar(new_fork, *i);
	  refmap[new_fork] = mxnode++;
	  //cerr<<" with new parent N"<<refmap[new_fork]<<"\n";
	}
      }
    }
  }

  //insert new_par btw ch & ch's predecessors
  void InsertPar(GRCNode *new_par, GRCNode *ch)
  {
    vector<GRCNode *>::iterator i,j;
    
    new_par->successors.push_back(ch);
    for(i = ch->predecessors.begin(); i != ch->predecessors.end(); i++){
      new_par->predecessors.push_back(*i);
      for (j = (*i)->successors.begin(); j != (*i)->successors.end(); j++){
	if ((*j) == ch)
	  *j = new_par;
      }
    }
    ch->predecessors.clear();
    ch->predecessors.push_back(new_par);
  }

  void BitmapPrint()
  {
    set<GRCNode *>::iterator i;
    set<GRCNode *>::iterator j;
    //set<Edge *>::iterator j;
    int t;
    //set<GRCNode *>::iterator k;

    for(i = visited.begin(); i != visited.end(); i++){
      cerr<<"N"<<refmap[*i]<<"\n";
      cerr<<"  reg: ";
      for (t = 0; t < pos_cnt; t++){
	if (region[*i]->GetBit(t) == 1)
	  for (j = visited.begin(); j != visited.end(); j++){
	    if (t == bitpos[*j]){
	      cerr<<refmap[*j]<<" ";
	      break;
	    }
	  }
      }
      cerr<<"\n";

      cerr<<"  eec: ";
      for (t = 0; t < pos_cnt; t++){
	if (eec[*i]->GetBit(t) == 1)
	  for (j = visited.begin(); j != visited.end(); j++){
	    if (t == bitpos[*j])
	      cerr<<refmap[*j]<<" ";
	  }
      }
      cerr<<"\n";
      
      /*cerr<<"exedges: ";
      
      for(j = relation[*i]->exedges.begin(); j != relation[*i]->exedges.end(); j++){
	cerr<<refmap[(*j)->begin]<<"->"<<refmap[(*j)->end]<<" ";
      }
      cerr<<"\n";

      cerr<<"descends: ";
      for (k = relation[*i]->descends.begin(); k != relation[*i]->descends.end(); k++){
	cerr<<refmap[*k]<<" ";
      }
      cerr<<"\n";*/
    }
  }

  void DFSRemoveForks(GRCNode *n)
  {
    vector<GRCNode *>::iterator i,j;
    
    if (visited.count(n)>0) 
      return;
    visited.insert(n);

    //if n is a non-NULL leaf, start[n] = ends[n] = n    
    if (conn_succ.find(n) == conn_succ.end()){
      start[n] = n;
      ends[n].push_back(n);
      return;
    }
    else if (conn_succ[n].size() == 0){
      start[n] = n;
      ends[n].push_back(n);
      return;
    }

    //if n is a fork, 
    //  start[n] = start[n->successors[0]], 
    //  ends[n] = ends[n->successors[size-1]]
    if (dynamic_cast<Fork *>(n)){

      DFSRemoveForks(*conn_succ[n].begin());

      for (i = conn_succ[n].begin()+1; i != conn_succ[n].end(); i++){

	DFSRemoveForks(*i);

	for (j = ends[*(i-1)].begin(); j != ends[*(i-1)].end(); j++){
	  Link(*j, start[*i]);
	}	
      }
      
      start[n] = start[*(conn_succ[n].begin())];
      ends[n] = ends[*(conn_succ[n].end()-1)];
    }
    else{
      start[n] = n;

      for (i = conn_succ[n].begin(); i != conn_succ[n].end(); i++){

	DFSRemoveForks(*i);
	Link(n,start[*i]);
	ends[n].insert(ends[n].end(),ends[*i].begin(), ends[*i].end());
      }
    }

    //cerr<<"finished N"<<refmap[n]<<" w/t start = "<<refmap[start[n]]<<"\n";
  }

  void RmLinks()
  {
    set<GRCNode *>::iterator i;
    for (i = visited.begin(); i != visited.end(); i++){
      (*i)->predecessors.clear();
      (*i)->successors.clear();
    }
  }

  void CheckAndLink(GRCNode *par, GRCNode *n)
  {
    vector<GRCNode *>::iterator i;
    bool found = false;

    if ((dynamic_cast<Switch *>(par)) 
	|| (dynamic_cast<Test *>(par)) 
	|| (dynamic_cast<Sync *>(par))){
    }
    else{// if not switch or test or sync, can't have multiple same child, dont add it
      for (i = par->successors.begin(); i != par->successors.end(); i++){
	if ((*i) == n){
	  found = true;
	  break;
	}
      }
    }
    if (!found){
      par->successors.push_back(n);
      n->predecessors.push_back(par);
    }	
  }
  
  void Link(GRCNode *par, GRCNode *n)
  {
    vector<GRCNode *>::iterator i;

    //the assumption is, the parent of par node must be a switch, test or sync
    // or it can't have null child in PDG
    if (dynamic_cast<Nop *>(par)){
      assert(par->predecessors.size() == 1);
      GRCNode *par_par = par->predecessors[0];
      for (i = par_par->successors.begin(); i != par_par->successors.end(); i++)
	if ((*i) == par){
	  *i = n;
	  break;
	}     
      par->predecessors.clear();
      n->predecessors.push_back(par_par);
    }
    else
      CheckAndLink(par, n);
  }

  void OrderChildren()
  {
    set<GRCNode *>::iterator i;
    vector<GRCNode *>::iterator x,y;
    bool inserted;

    for (i = visited.begin(); i != visited.end(); i++){      
      //cerr<<"For N"<<refmap[*i]<<":\n";
      if ((dynamic_cast<Fork *>(*i)) && ((*i)->successors.size()>1)){

	vector<GRCNode *> ordered;
	map<GRCNode *, SignalSymbol *> cuthead;

	//cerr<<"A Fork N"<<refmap[*i]<<":-----------\n";
	
	ordered.push_back(*((*i)->successors.begin()));
	for (x = (*i)->successors.begin()+1; x != (*i)->successors.end(); x++){

	  inserted = false;

	  for (y = ordered.begin(); y != ordered.end(); y++){
	    //cerr<<"comp N"<<refmap[*x]<<"-N"<<refmap[*y]<<"\n";
	    switch (OrderCheck(*x,*y)){
	    case 1://-
	      //cerr<<"c1\n";
	      if (DPSOn(*y,*x)){
		y = ordered.insert(y,*x); 
		inserted = true; //set done
	      }
	      break;
	    case 2:// y < x
	      //cerr<<"c2\n";
	      if (DPSOn(*y,*x)){ //a cut on y
		CutVarSet(*y,cuthead);
		y = ordered.insert(y,*x);
		inserted = true;
	      }
	      //else, insert it later
	      break;
	    case 3:// y > x
	      //cerr<<"c3\n";
	      if (DPSOn(*x, *y)){
		CutVarSet(*x, cuthead);//make a cut but insert it later
	      }
	      else{
		y = ordered.insert(y,*x); 
		inserted = true;
	      }
	      break;
	    case 4:// y <> x
	      //cerr<<"c4\n";
	      if (DPSOn(*y,*x)){ // a cut on y
		CutVarSet(*y, cuthead);
		y = ordered.insert(y,*x);
		inserted = true;
	      }
	      else { // a cut on x but insert it later
		CutVarSet(*x, cuthead);
	      }
	      break;
	    case 5:
	      assert(0);
	      break;
	    }
	    if (inserted)
	      break;
	  }
	  if(!inserted)
	    ordered.push_back(*x);
	}

	assert(ordered.size() == (*i)->successors.size());
	bool cutted = false;
	for (x = ordered.begin(); x != ordered.end(); x++){
	  // if x not need be cutted and no node ordered bf has been cutted
	  if ((cuthead.count(*x) == 0) && (!cutted)){	 
	    conn_succ[*i].push_back(*x);
	    conn_pred[*x].push_back(*i);
	  }
	  else {
	    cutted = true;
	    if(cuthead.count(*x) == 0){
	      SignalSymbol *new_s = NewVariable();
	      sigs->enter(new_s);
	      cuthead[*x] = new_s;
	    }
	    GRCNode *new_set_s = NewSetVar(cuthead[*x],1);
	    ordered.insert(ordered.begin(), new_set_s);
	    GRCNode *new_test_s = NewTester(cuthead[*x]);
	    conn_succ[new_test_s].push_back(NULL);
	    conn_succ[new_test_s].push_back(*x);
	    conn_pred[*x].push_back(new_test_s);

	    conn_succ[*i].push_back(new_test_s);
	    conn_pred[new_test_s].push_back(*i);
	  }
	}
      }
      else{
	for (x = (*i)->successors.begin(); x != (*i)->successors.end(); x++){
	  conn_succ[*i].push_back(*x);
	  conn_pred[*x].push_back(*i);
	}
      }
    }
  }

  void CutVarSet(GRCNode *n, map<GRCNode *, SignalSymbol *> &cuthead)
  {
    if (cuthead.count(n) == 0) {//never been cutted before
      SignalSymbol *new_s = NewVariable();
      sigs->enter(new_s);
      cuthead[n] = new_s;
    }
  }

  GRCNode *NewSetVar(SignalSymbol *s, int val)  
  {
    Emit *em = new Emit(s, NULL);
    return (new Action(em));
  }

  GRCNode *NewTester(SignalSymbol *s)
  {
    // FIXME: should be the built-in boolean type
    LoadSignalExpression *lde = new LoadSignalExpression(NULL, s);
    return (new Test(NULL, lde));
  }

  //FIXME: it's signal currently, should be changed to varialbe
  SignalSymbol *NewVariable()
  {
    char new_sig_name[32];
    char *pref = "_VAR_G";

    cerr<<"A CUT!!!!!!!!!!!!!!!!\n";

    sprintf(new_sig_name, "%s%d", pref, var_cnt);
    var_cnt++;
    //FIXME: later should be a separate table to store thess new signals
    return (new SignalSymbol(new_sig_name,NULL, SignalSymbol::Input,NULL,NULL));
  }

  //return the order type of two nodes: 1 -, 2 Y<X, 3 Y>X, 4 Y<>X, 5 N/A
  int OrderCheck(GRCNode *x, GRCNode *y)
  {
    switch (TypeCheck(y)){
    case 1:
      switch (TypeCheck(x)){
      case 1:
	return 1;
      case 2:
	if (eec[x]->GetBit(bitpos[x]) == 1)
	  return 1;
	return 2;
      case 3:
	if (eec[y]->GetBit(bitpos[x]) == 1)
	  return 2;
	return 5;
      }
      break;
    case 2:
      switch (TypeCheck(x)){
      case 1:
	if (eec[y]->GetBit(bitpos[y]) == 1)
	  return 1;
	return 3;
      case 2:
	if ((eec[y]->GetBit(bitpos[y]) == 1) 
	    && (eec[x]->GetBit(bitpos[x]) == 1))
	  return 1;
	if (eec[y]->GetBit(bitpos[y]) == 1)
	  return 2;
	if (eec[x]->GetBit(bitpos[x]) == 1)
	  return 3;
	//cerr<<"here41\n";
	return 4;
      case 3:
	if (eec[y]->GetBit(bitpos[x]) == 1)
	  return 2;
	if (eec[y]->GetBit(bitpos[y]) == 1)
	  return 5;
	//cerr<<"here42\n";
	return 4;
      }
      break;
    case 3:
      switch (TypeCheck(x)){
      case 1:
	if (eec[x]->GetBit(bitpos[y]) == 1)
	  return 3;
	return 5;
      case 2:
	if (eec[x]->GetBit(bitpos[y]) == 1)
	  return 3;
	if (eec[x]->GetBit(bitpos[x]) == 1)
	  return 5;
	//cerr<<"here43\n";
	return 4;
      case 3:
	if (eec[x]->GetBit(bitpos[y]) == 1) {
	  if (eec[y]->GetBit(bitpos[x]) == 1)
	    return 1;
	  else return 3;
	} 
	if (eec[y]->GetBit(bitpos[x]) == 1)
	  return 2;
	//cerr<<"here44\n";
	return 4;
      }
    }
    assert(0);//unknow case
    return 5;
  }
    
  //test if x has data dependence on y: y--->x
  bool DPSOn(GRCNode *x, GRCNode *y)
  {
    set<GRCNode *>::iterator i;
    vector<GRCNode *>::iterator j;

    for (i = dpsnodes[x].begin(); i != dpsnodes[x].end(); i++){
      for (j = (*i)->dataPredecessors.begin();
	   j != (*i)->dataPredecessors.end(); j++)
	if (dpsnodes[y].count(*j)> 0)
	  return true;
    }
    return false;
  }

  void ComputeRegion()
  {
    vector<GRCNode *>::iterator i,j,k;
    BITMAP *par_reg;

    for (i = topo.begin(); i != topo.end(); i++){
      //cerr<<"N"<<refmap[*i]<<"\n";
      //relation[*i]= new ExEdge(); //need later
      if ((*i) == entergrc){
	region[*i] = new BITMAP(pos_cnt,0);
      }
      else {
	assert((*i)->predecessors.size()>0);
	BITMAP *RET = new BITMAP(pos_cnt,1);
	//cerr<<" init ";RET->print();
	for (j = (*i)->predecessors.begin(); j != (*i)->predecessors.end(); j++){	  
	  //cerr<<" AND par "<<refmap[*j]<<"\n";
	  assert(region.count(*j)>0);
	  if(dynamic_cast<Fork *>(*j)){
	    par_reg = region[*j]->copy();
	    for (k = (*j)->successors.begin(); k != (*j)->successors.end(); k++){
	      par_reg->SetBit(bitpos[*k],1);
	    }
	    //cerr<<"its a fork\n";
	  }
	  else {
	    par_reg = region[*j];
	    //cerr<<"its NOT a fork\n";
	  }
	  //cerr<<"par "<<refmap[*j]<<"->"<<refmap[*i]<<" ";par_reg->print();
	  RET->And(par_reg);
	  //cerr<<"aft AND    ";RET->print();
	}
	region[*i] = RET;   
	//cerr<<"region "<<refmap[*i]<<"  ";region[*i]->print();
      }
    }
  }

  void DFSEec(GRCNode *n)  
  {
    if(!n) return;
    if(visited.count(n)>0) return;
    visited.insert(n);
    
    //relation[n]->AddDesc(n);
    if (n->successors.size() == 0){
      eec[n] = region[n];
      return;
    }

    BITMAP *tmp = new BITMAP(pos_cnt,1);
    for (vector<GRCNode *>::iterator i = n->successors.begin();
	 i != n->successors.end(); i++){
      if (*i){
	DFSEec(*i);
	//relation[n]->AddDesc(relation[*i]->descends);
	
	assert(eec.count(*i)>0);
	tmp->And(eec[*i]);
      }
    }
    eec[n] = tmp;
    //relation[n]->FindExedges(n);
  }

  int TypeCheck(GRCNode *n)
  {
    if ((dynamic_cast<Test *>(n))
	|| (dynamic_cast<Sync *>(n)) 
	|| (dynamic_cast<Switch *>(n)))
      return 2;

    if (dynamic_cast<Fork *>(n))
      return 3;

    //assert(n->successors.size() == 0);
    if(n->successors.size() != 0)
      assert(dynamic_cast<EnterGRC *>(n));
    return 1;
  }
};

int main(int argc, char *argv[])
{

  if (argc < 4) {
    cerr<<"Usage: ... ccfg_file scfg_file dot_file\n"; return -1;}

  ifstream sf_in(argv[1]); if(!sf_in.good()) return -1;
  ofstream sf_out(argv[2]); if(!sf_out.good()) return -1;
  ofstream df_out(argv[3]); if(!df_out.good()) return -1;

  try {
    IR::XMListream r(sf_in);
    IR::Node *n;
    r >> n;
    sf_in.close();
    AST::Modules *mods = dynamic_cast<AST::Modules*>(n);
    if (!mods) throw IR::Error("Root node is not a Modules object");

    for ( vector<AST::Module*>::iterator i = mods->modules.begin() ;
	  i != mods->modules.end() ; i++ ) {

      assert(*i);

      GRCgraph *gf = dynamic_cast<GRCgraph*>((*i)->body);
      assert(gf);  
      GRCNode *top = gf->control_flow_graph;

      AST::SymbolTable *sig = dynamic_cast<AST::SymbolTable*>((*i)->signals);
      assert(sig);

      CFGmap refmap;
      STmap strefmap;
      
      int mxnode = gf->enumerate(refmap,strefmap);

      EEC eec_computer(top, sig, refmap, mxnode);
      eec_computer.excute();

      //remove EnterGRC
      //gf->control_flow_graph = eec_computer.rtTop();
      //assert(gf->control_flow_graph->successors.size()>0);

      mxnode = GRCDot::GRCDot(df_out, gf,(*i),false, false, refmap, strefmap, mxnode);

      df_out.close();

    }
    
    IR::XMLostream o(sf_out);
    o << n;
    sf_out.close();
  } catch (IR::Error &e) {
    cerr << e.s << endl;
    exit(-1);
  }

  return 0;

}
